/*
 * active port forwarder - software for secure forwarding
 * Copyright (C) 2003 jeremian <jeremian@poczta.fm>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "activefor.h"
#include "network.h"

#include <openssl/ssl.h>

static void usage(char* info);

int
main(int argc, char **argv)
{
	int	masterfd, i, n, numofcon, length;
	ConnectuserT* contable = NULL;
	unsigned char				buff[8096];
	char hostname[100];
	int			maxfdp1, usernum, usercon;
	fd_set		rset, allset;
	char verbose = 0;
	char* name = NULL;
	char* manage = NULL;
	char* desnam = NULL;
	char* despor = NULL;
	char* keys = NULL;

	SSL_METHOD* method;
	SSL_CTX* ctx;
	SSL* ssl;

	while ((n = getopt(argc, argv, "hn:m:d:p:vk:")) != -1) {
		switch (n) {
		  case 'h': {
				    usage("Active port forwarder (client) v0.5");
					  break;
				  }
		  case 'n': {
				    name = optarg;
				    break;
			    }
		  case 'm': {
				    manage = optarg;
				    break;
			    }
		  case 'd': {
				    desnam = optarg;
				    break;
			    }
		  case 'p': {
				    despor = optarg;
				    break;
			    }
		  case 'v': {
				    verbose = 1;
				    break;
			    }
		  case 'k': {
				    keys = optarg;
				    break;
			    }
		  case '?': {
				    usage("Unrecognized option");
				    break;
			    }
		}
	}

	if (name == NULL) {
		usage("Name of the server is required");
	}
	if (manage == NULL) {
		manage = "50126";
	}
	if (desnam == NULL) {
		gethostname(hostname, 100);
		desnam = hostname;
	}
	if (despor == NULL) {
		usage("Destination port number is required");
	}
	if (keys == NULL) {
		keys = "client.rsa";
	}


	masterfd = tcp_connect(name, manage);

	
	SSL_library_init();
	method = SSLv3_client_method();
	ctx = SSL_CTX_new(method);
	if (SSL_CTX_set_cipher_list(ctx, "ALL:@STRENGTH") == 0) {
		printf("Setting cipher list failed... exiting\n");
		exit(1);
	}
	if (SSL_CTX_use_RSAPrivateKey_file(ctx, keys, SSL_FILETYPE_PEM) != 1) {
		printf("Setting rsa key failed (%s)... exiting\n", keys);
		exit(1);
	}
	ssl = SSL_new(ctx);
	if (SSL_set_fd(ssl, masterfd) != 1) {
		printf("Problem with initializing ssl... exiting\n");
		exit(1);
	}
	if (verbose)
		printf("Trying SSL_connect\n");
	if ((n = SSL_connect(ssl)) == 1) {
		if (verbose) {
			printf("SSL_connect successfull\n");
		}
	}
	else {
		printf("SSL_connect has failed (%d)... exiting\n", n);
		exit(1);
	}

	
	buff[0] = AF_S_LOGIN;
	buff[1] = 1;
	buff[2] = 3;
	buff[3] = 6;
	buff[4] = 2;
	SSL_write(ssl, buff, 5);
	buff[0] = 0;
	SSL_read(ssl, buff, 5);
	
	if ( buff[0] != AF_S_LOGIN ) {
		printf("Incompatible server type or server full\n");
		exit(1);
	}

	usernum = buff[1];
	usernum = usernum << 8;
	usernum += buff[2];
	
	contable = calloc( usernum, sizeof(ConnectuserT));
	if (contable == NULL) {
		printf("Calloc error - unable to succesfully comunicate with server\n");
		exit(1);
	}

	if (!verbose)
		daemon(0, 0);
	
	FD_ZERO(&allset);
	
	FD_SET(masterfd, &allset);
	maxfdp1 = masterfd + 1;
	
	
	for ( ; ; ) {
		rset = allset;
		if (verbose)
			printf(">select\n");
		select(maxfdp1, &rset, NULL, NULL, NULL);
		if (verbose)
			printf(">>after select...\n");

		for (i = 0; i < usernum; ++i) {
			if (FD_ISSET(contable[i].connfd, &rset)) {	
				if (verbose)
					printf("FD_ISSET(contable[%d].connfd)\n", i);
				n = read(contable[i].connfd, &buff[5], 8091);
				if (n) {
					buff[0] = AF_S_MESSAGE; /* sending message */
					buff[1] = i >> 8;	/* high bits of user number */
					buff[2] = i;		/* low bits of user number */
					buff[3] = n >> 8;	/* high bits of message length */
					buff[4] = n;		/* low bits of message length */
					SSL_write(ssl, buff, n+5);
				}
				else {
					close(contable[i].connfd);
					FD_CLR(contable[i].connfd, &allset);
					contable[i].state = S_STATE_CLOSING;
					buff[0] = AF_S_CONCLOSED; /* closing connection */
					buff[1] = i >> 8;	/* high bits of user number */
					buff[2] = i;		/* low bits of user number */
					SSL_write(ssl, buff, 5);
				}
			}
		}

		if (FD_ISSET(masterfd, &rset)) {
			if (verbose)
				printf("FD_ISSET(masterfd)\n");
			n = SSL_read(ssl, buff, 5);
			if (n == 0) { /* server quits -> we do the same... */
				printf("premature quit of the server -> exiting...\n");
				exit(1);
			}
			numofcon = buff[1];
			numofcon = numofcon << 8;
			numofcon += buff[2]; /* this is id of user */
			length = buff[3];
			length = length << 8;
			length += buff[4]; /* this is length of message */
			switch (buff[0]) {
				case AF_S_CONCLOSED : {
						      if ((numofcon>=0) && (numofcon<=usernum)) {
							      usercon--;
							      if (contable[numofcon].state == S_STATE_CLOSING) {
								      contable[numofcon].state = S_STATE_CLEAR;
							      }
							      else if (contable[numofcon].state==S_STATE_OPEN) {
					      close(contable[numofcon].connfd);
					      FD_CLR(contable[numofcon].connfd, &allset);
					      contable[numofcon].state = S_STATE_CLEAR;
					      buff[0] = AF_S_CONCLOSED; /* closing connection */
					      buff[1] = numofcon >> 8;		/* high bits of user number */
					      buff[2] = numofcon;		/* low bits of user number */
					      SSL_write(ssl, buff, 5);
							      }
						      }
						      break;
						      }
				case AF_S_CONOPEN : {
						if ((numofcon>=0) && (numofcon<=usernum)) {
							usercon++;
 						  if (contable[numofcon].state == S_STATE_CLEAR) {
						    contable[numofcon].connfd = tcp_connect(desnam, despor);
						    FD_SET(contable[numofcon].connfd, &allset);
						    maxfdp1 = (maxfdp1 > (contable[numofcon].connfd+1)) ? maxfdp1 : (contable[numofcon].connfd+1);
					    buff[0] = AF_S_CONOPEN; /* closing connection */
					    buff[1] = numofcon >> 8;		/* high bits of user number */
					    buff[2] = numofcon; 		/* low bits of user number */
					    SSL_write(ssl, buff, 5);
					    contable[numofcon].state = S_STATE_OPEN;
						  }
						}
						break;
						    }
				case AF_S_MESSAGE : {
							    printf("Received msg for con[%d]\n", numofcon);
							    n = SSL_read(ssl, buff, length);
							    if (n != length) {
							      if (verbose) {
								printf("n(%d)!=length(%d)\n", n, length);
							      }
							      break;
							    }
						if ((numofcon>=0) && (numofcon<=usernum)) {
							if (contable[numofcon].state == S_STATE_OPEN) {
							if (verbose)
							printf("sent msg con[%d], length=%d\n", numofcon, n);
							  write(contable[numofcon].connfd, buff, n);
							}
						}
							    break;
						    }
				default : { /* unrecognized type of message -> exiting... */
					  printf("Server sents unrecognized message -> exiting...\n");
					  exit(1);
						  
						  break;
					  }
			}
		}
	}
}

static void
usage(char* info)
{
    printf("\n%s\n\n", info);
    printf("  Options:\n");
    printf("  -h                      - prints this help\n");
    printf("  -n [server name]        - where the second part of the active\n");
    printf("                            port forwarder is running (required)\n");
    printf("  -m [portnum]            - the manage port number - server must\n");
    printf("                            listening on it (default: 50126)\n");
    printf("  -d [hostname]           - name of this host/remote host - the final\n");
    printf("                            destination of the packets (default: name\n");
    printf("                            returned by hostname function)\n");
    printf("  -p [portnum]            - port we are forwarding connection to (required)\n");
    printf("  -k [keyfile]            - name of the file with RSA key (default: client.rsa)\n");
    printf("  -v                      - to be verbose - program won't enter into\n");
    printf("                            the daemon mode\n\n");
    exit(0);
}

